/*
 * Generated by asn1c-0.9.27 (http://lionet.info/asn1c)
 * From ASN.1 module "GssapiSessionModule"
 * 	found in "session.asn1"
 */

#ifndef	_Uint32_H_
#define	_Uint32_H_


#include <asn_application.h>

/* Including external dependencies */
#include <NativeInteger.h>

#ifdef __cplusplus
extern "C" {
#endif

/* Uint32 */
typedef unsigned long	 Uint32_t;

/* Implementation */
extern asn_TYPE_descriptor_t asn_DEF_Uint32;
asn_struct_free_f Uint32_free;
asn_struct_print_f Uint32_print;
asn_constr_check_f Uint32_constraint;
ber_type_decoder_f Uint32_decode_ber;
der_type_encoder_f Uint32_encode_der;
xer_type_decoder_f Uint32_decode_xer;
xer_type_encoder_f Uint32_encode_xer;

#ifdef __cplusplus
}
#endif

#endif	/* _Uint32_H_ */
#include <asn_internal.h>
